import pathlib, pytest
import numpy as np
from numpy.testing import assert_array_equal
import pandas as pd
import seaborn as sns

import sklearn
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.datasets import make_classification, make_regression
from sklearn.metrics import accuracy_score, roc_auc_score
from sklearn.preprocessing import StandardScaler, OneHotEncoder

from mercury.dataschema import DataSchema

from mercury.robust.model_tests import (
    ModelReproducibilityTest,
    ModelSimplicityChecker,
    DriftPredictionsResistanceTest,
    ClassificationInvarianceTest,
    TreeCoverageTest,
    DriftMetricResistanceTest,
    FeatureCheckerTest
)
from mercury.robust.errors import FailedTestError


@pytest.fixture(scope="module")
def datasets():
    tips = sns.load_dataset("tips")
    tips["sex"] = tips["sex"].astype(str)
    tips["smoker"] = tips["smoker"].astype(str)
    tips["day"] = tips["day"].astype(str)
    tips["time"] = tips["time"].astype(str)

    titanic = sns.load_dataset("titanic")
    isna_deck = titanic.deck.isna()
    titanic["class"] = titanic["class"].astype(str)
    titanic["deck"] = titanic["deck"].astype(str)
    titanic["who"] = titanic["who"].astype(str)
    titanic.loc[isna_deck, "deck"] = np.nan

    return tips, titanic


# TEST MODEL REPRODUCIBILITY TEST

@pytest.fixture(scope='module')
def classification_dataset():
    # Create Dataset for the model reproducibility test
    X,y = make_classification(n_samples=2000, n_informative=5, flip_y=0.5, random_state=10)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=10)

    df_train = pd.DataFrame(X_train, columns=["x" + str(i) for i in range(X_train.shape[1])])
    df_train["y"] = y_train

    df_test = pd.DataFrame(X_test, columns=["x" + str(i) for i in range(X_test.shape[1])])
    df_test["y"] = y_test

    return df_train, df_test

class CustomModel:

    def __init__(self):
        self.m1 = RandomForestClassifier(max_depth=4, random_state=4)
        self.m2 = RandomForestClassifier(random_state=3)
        self.m3 = RandomForestClassifier(max_depth=3, random_state=2)

    def fit(self, X,y):
        self.m1 = self.m1.fit(X,y)
        self.m2 = self.m2.fit(X,y)
        self.m3 = self.m3.fit(X,y)

    def predict(self, X):
        pred_m1 = self.m1.predict(X).reshape(-1,1)
        pred_m2 = self.m2.predict(X).reshape(-1,1)
        pred_m3 = self.m3.predict(X).reshape(-1,1)
        predictions = np.concatenate([pred_m1, pred_m2, pred_m3], axis=1)
        return np.median(predictions, axis=1)

    def __deepcopy__(self, memo):
        m = CustomModel()
        m.m1 = sklearn.base.clone(self.m1)
        m.m2 = sklearn.base.clone(self.m2)
        m.m3 = sklearn.base.clone(self.m3)
        return m

class NotValidModel:

    def __init__(self):
        pass

    def fit(self, X,y):
        pass

    def predict(self, X):
        return np.ones((X.shape[0]))


def test_model_reproducibility_sklearn(classification_dataset):

    df_train, df_test = classification_dataset

    model_repro = RandomForestClassifier(max_depth=5, random_state=2000)
    model_no_repro = RandomForestClassifier(max_depth=5, random_state=None)

    def train_model(model, X, y, train_params=None):
        model.fit(X,y)
        return model

    def eval_model(model, X, y, eval_params=None):
        y_pred = model.predict(X)
        return accuracy_score(y, y_pred)

    def get_predictions(model, X, predict_params=None):
        return model.predict(X)

    # Test Reproducible sklearn Model - Should pass
    test = ModelReproducibilityTest(
        model = model_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    test.run()

    # Test non-reproducible sklearn Model - Should fail(Diff Metric)
    test = ModelReproducibilityTest(
        model = model_no_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threhsold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()

    # Test non-reproducible sklearn model - Should fail (Diff predictions)
    test = ModelReproducibilityTest(
        model = model_no_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = None,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()

    # If we don't specify train function, an exception must be raised
    test = ModelReproducibilityTest(
        model = model_repro,
        train_dataset = df_train,
        target="y",
        train_fn = None,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    with pytest.raises(ValueError) as exinfo:
        test.run()

    # If we don't specify the eval function neither the predict function, an expection must be raised
    test = ModelReproducibilityTest(
        model = model_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = None,
        train_params = None,
        eval_params = None,
        predict_fn = None,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    with pytest.raises(ValueError) as exinfo:
        test.run()

    # Only check the eval metric - This Test should pass
    test = ModelReproducibilityTest(
        model = model_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = None,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    test.run()

    # Only Check a a training Set - This Test should pass
    test = ModelReproducibilityTest(
        model = model_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0
    )
    test.run()

    # Setting a high threshold_eval can make a non-exactly reproducible model pass the test
    test = ModelReproducibilityTest(
        model = model_no_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = None,
        threshold_eval = 100,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    test.run()

    # Setting a high threshold_yhat can make a non-exactly reproducible model pass the test
    test = ModelReproducibilityTest(
        model = model_no_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = None,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 100,
        test_dataset = df_test
    )
    test.run()

    # Similarly, we can set yhat_allowed_diff to allow a non-exactly reproducible model pass
    test = ModelReproducibilityTest(
        model = model_no_repro,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = None,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test,
        yhat_allowed_diff = 100
    )
    test.run()

def test_model_reproducibility_tfkeras(classification_dataset):

    df_train, df_test = classification_dataset

    from tensorflow.keras.layers import Dense, Input
    from tensorflow.keras import Sequential
    model = Sequential()
    model.add(Input(shape=(df_train.shape[1]-1,)))
    model.add(Dense(10, activation='relu'))
    model.add(Dense(2, activation='softmax'))

    def train_model(model, X, y, train_params):
        model.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        hist = model.fit(X, y, batch_size=train_params["batch_size"], epochs=train_params["epochs"])
        return model

    def eval_model(model, X, y, eval_params=None):
        y_pred = np.argmax(model.predict(X), axis=1)
        return accuracy_score(y, y_pred)

    def get_predictions(model, X):
        return np.argmax(model.predict(X), axis=1)

    # Model is not reproducible - this test should fail
    test = ModelReproducibilityTest(
        model = model,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = {"batch_size": 8, "epochs": 2},
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()


def test_model_reproducibility_custom_model(classification_dataset):

    df_train, df_test = classification_dataset

    # TEST CUSTOM MODEL
    m = CustomModel()
    def train_model(model, X, y, train_params=None):
        model.fit(X,y)
        return model

    def eval_model(model, X, y, eval_params=None):
        y_pred = model.predict(X)
        return accuracy_score(y, y_pred)

    def get_predictions(model, X, predict_params=None):
        return model.predict(X)

    # The model is reproducible - The Test should Pass
    test = ModelReproducibilityTest(
        model = m,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    test.run()


    # TEST CUSTOM MODEL NOT COMPATIBLE WITH REPRODUCIBLITY TEST
    m = NotValidModel()

    def train_model(model, X, y, hyperparams=None):
        model.fit(X,y)
        return model

    def eval_model(model, X, y, eval_params=None):
        y_pred = model.predict(X)
        return accuracy_score(y, y_pred)

    def get_predictions(model, X):
        return model.predict(X)

    # The model is not compatible with the test - An exception must be raised
    test = ModelReproducibilityTest(
        model = m,
        train_dataset = df_train,
        target="y",
        train_fn = train_model,
        eval_fn = eval_model,
        train_params = None,
        eval_params = None,
        predict_fn = get_predictions,
        threshold_eval = 0,
        threshold_yhat = 0,
        test_dataset = df_test
    )
    with pytest.raises(ValueError) as exinfo:
        test.run()


# MODEL SIMPLICITY CHECKER

def _create_classification_pipeline(num_feats=None, cat_feats=None):

    numeric_transformer = Pipeline(steps=[("scaler", StandardScaler())])
    categorical_transformer = OneHotEncoder(handle_unknown="ignore")

    preprocessor = ColumnTransformer(
        transformers=[
            ("num", numeric_transformer, num_feats),
            ("cat", categorical_transformer, cat_feats),
        ], remainder='drop'
    )

    pipeline = Pipeline(
        steps=[("preprocessor", preprocessor),
               ("classifier", RandomForestClassifier(n_estimators=200, max_depth=100, random_state=2000))]
    )

    return pipeline

def _create_regression_pipeline(num_feats=None, cat_feats=None):

    numeric_transformer = Pipeline(steps=[("scaler", StandardScaler())])
    categorical_transformer = OneHotEncoder(handle_unknown="ignore")

    preprocessor = ColumnTransformer(
        transformers=[
            ("num", numeric_transformer, num_feats),
            ("cat", categorical_transformer, cat_feats),
        ], remainder='drop'
    )

    pipeline = Pipeline(
        steps=[("preprocessor", preprocessor),
               ("classifier", RandomForestRegressor(n_estimators=200, max_depth=100, random_state=2000))]
    )

    return pipeline

def test_model_simplicity_checker_sk_classification(datasets):

    tips, titanic = datasets
    titanic2 = titanic.copy()
    titanic2 = titanic2.drop(['deck', 'alive'], axis=1).dropna()

    target = 'survived'
    num_feats = ['age', 'fare']
    cat_feats = ['pclass', 'sex', 'sibsp', 'parch', 'embarked', 'class', 'who', 'adult_male', 'embark_town', 'alone']

    df_train, df_test = train_test_split(titanic2, test_size=0.3, random_state=2000)
    X_train = df_train.loc[:, [f for f in df_train.columns if f != target]]
    y_train = df_train.loc[:, target]
    X_test = df_test.loc[:, [f for f in df_test.columns if f != target]]
    y_test = df_test.loc[:, target]

    # create and train Model
    model = _create_classification_pipeline(num_feats, cat_feats)
    model = model.fit(X_train, y_train)

    # With LogReg as baseline, the test fails
    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        threshold=0.01,
        baseline_model=LogisticRegression(solver='liblinear')
    )
    assert test.info() is None
    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert (test.metric_model - 0.01) <= test.metric_baseline_model
    assert 'metric_model' in test.info()

    # With DecTree as baseline, the test passes
    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        threshold=0.01,
        baseline_model=DecisionTreeClassifier(random_state=2000),
        task='classification'  # task not inferred in this case
    )
    test.run()
    assert (test.metric_model - 0.01) > test.metric_baseline_model

    # Run with ignore_feats
    X_train["random_feature"] = np.random.uniform(size=len(X_train))
    X_test["random_feature"] = np.random.uniform(size=len(X_test))
    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        ignore_feats=["random_feature"],
        threshold=0.01,
        baseline_model=DecisionTreeClassifier(random_state=2000),
        task='classification'  # task not inferred in this case
    )
    test.run()

def test_model_simplicity_checker_sk_regression(datasets):

    tips, titanic = datasets

    target = 'tip'
    num_feats = ['total_bill']
    cat_feats = ['sex', 'smoker', 'day', 'time', 'size']

    df_train, df_test = train_test_split(tips, test_size=0.3, random_state=2000)
    X_train = df_train.loc[:, [f for f in df_train.columns if f != target]]
    y_train = df_train.loc[:, target]
    X_test = df_test.loc[:, [f for f in df_test.columns if f != target]]
    y_test = df_test.loc[:, target]

    # create and train the model
    model = _create_regression_pipeline(num_feats, cat_feats)
    model = model.fit(X_train, y_train)

    # test with default baseline
    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        threshold=0.02
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert (test.metric_model - 0.01) <= test.metric_baseline_model

def test_model_simplicity_checker_tfkeras():

    # create dataset
    X,y = make_classification(n_samples=2000, n_informative=5, flip_y=0.5, random_state=10)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=10)

    # Create tf.keras model and train
    import tensorflow as tf
    from tensorflow.keras import layers
    from tensorflow.keras.layers import Dense, Input

    model = tf.keras.Sequential([
        Input(shape=(X_train.shape[1],)),
        layers.Dense(30, input_dim=8, activation='relu'),
        layers.Dense(20, activation='relu'),
        layers.Dense(2, activation='softmax',  name='output')
    ])
    model.compile(optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
    history = model.fit(X_train, y_train, epochs=5)

    # Test using custom predict_fn
    def custom_predict(model, X):
        y_hat = np.argmax(model.predict(X), axis=1)
        return y_hat

    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        threshold=0.1,  # high threhsold to ensure randomness doesn't make different outcome in executions
        predict_fn=custom_predict
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert (test.metric_model - 0.1) <= test.metric_baseline_model

    # Test with already given predictions and eval_fn
    # Calculate predictions
    y_hat_test = custom_predict(model, X_test)
    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        threshold=0.1,  # high threhsold to ensure randomness doesn't make different outcome in executions
        eval_fn=roc_auc_score,
        test_predictions=y_hat_test,
        encode_cat_feats=False,
        scale_num_feats=False
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert (test.metric_model - 0.1) <= test.metric_baseline_model

def test_model_simplicity_checker_errors(datasets):

    tips, titanic = datasets

    target = 'tip'
    num_feats = ['total_bill']
    cat_feats = ['sex', 'smoker', 'day', 'time', 'size']

    df_train, df_test = train_test_split(tips, test_size=0.3, random_state=2000)
    X_train = df_train.loc[:, [f for f in df_train.columns if f != target]]
    y_train = df_train.loc[:, target]
    X_test = df_test.loc[:, [f for f in df_test.columns if f != target]]
    y_test = df_test.loc[:, target]

    # create and train the model
    model = _create_regression_pipeline(num_feats, cat_feats)
    model = model.fit(X_train, y_train)

    # Wrong task raises exception
    test = ModelSimplicityChecker(
        model = model,
        X_train = X_train,
        y_train = y_train,
        X_test = X_test,
        y_test = y_test,
        threshold=0.02,
        task='image_classification'
    )
    with pytest.raises(ValueError) as exinfo:
        test.run()

    # Wrong target type (datetime)
    X = np.array([1,2,3])
    y = np.array([np.datetime64('2005-02-25'), np.datetime64('2005-02-25'), np.datetime64('2005-02-25')])
    test = ModelSimplicityChecker(
        model = None,
        X_train = X,
        y_train = y,
        X_test = X,
        y_test = y,
        threshold=0.1,  # high threhsold to ensure randomness doesn't make different outcome in executions
    )
    with pytest.raises(ValueError) as exinfo:
        test.run()

def test_model_simplicity_multiclass_classification():

    # Generate multiclass classification dataset
    X, y = make_classification(n_samples=200, n_features=10, n_informative=5, n_classes=3)
    y_ohe = np.zeros((y.size, y.max()+1))
    y_ohe[np.arange(y.size),y] = 1

    # Test 1: Labels encoded as labels
    m1 = RandomForestClassifier()
    m1.fit(X, y)
    test = ModelSimplicityChecker(
        model = m1,
        X_train = X.copy(),
        y_train = y.copy(),
        X_test = X.copy(),
        y_test = y.copy(),
        threshold=0.01,
    )
    test.run()
    assert (test.metric_model - 0.01) > test.metric_baseline_model

    # Test 2: Lbaels encoded as ohe
    m2 = RandomForestClassifier()
    m2.fit(X, y_ohe)
    test = ModelSimplicityChecker(
        model = m2,
        X_train = X.copy(),
        y_train = y_ohe.copy(),
        X_test = X.copy(),
        y_test = y_ohe.copy(),
        baseline_model =  DecisionTreeClassifier(max_depth=1),
        threshold=0.01,
    )
    test.run()
    assert (test.metric_model - 0.01) > test.metric_baseline_model


def test_drift_prediction_resistance_test():
    df = pd.read_csv("tutorials/data/credit/UCI_Credit_card.csv")

    df = df.loc[:, df.columns != 'ID']

    col_keys = list(df.columns)
    col_vals = [x.lower() if x != 'default.payment.next.month' else 'Y' for x in col_keys]

    df = df.rename(columns = dict(zip(col_keys, col_vals)))

    X = df.loc[:, df.columns != 'Y'].copy()
    Y = df.loc[:, 'Y']

    rf = RandomForestClassifier()

    model = rf.fit(X, Y)

    drift_args = {'cols' : ['pay_amt1', 'pay_amt2', 'pay_amt3', 'pay_amt4', 'pay_amt5', 'pay_amt6'], 'iqr' : [1, 1.1]}
    test = DriftPredictionsResistanceTest(model = model, X = X, drift_type = 'scale_drift', drift_args = drift_args, tolerance = X.shape[0]*0.05)
    test.run()
    assert test.info()["loss"] < test.tolerance

    drift_args = {'force' : 200, 'noise' : 100}
    test = DriftPredictionsResistanceTest(model			 = model,
                               X				 = X,
                               drift_type		 = 'shift_drift',
                               drift_args		 = drift_args,
                               names_categorical = ['sex', 'education', 'marriage'],
                               tolerance		 = X.shape[0]*0.05)

    with pytest.raises(FailedTestError):
        test.run()

    drift_args = {'cols' : ['pay_amt1', 'pay_amt2', 'pay_amt3', 'pay_amt4', 'pay_amt5', 'pay_amt6'], 'iqr' : [1, 1.1]}

    with pytest.raises(RuntimeError):
        test = DriftPredictionsResistanceTest(model = model, X = X, drift_type = 'not_a_method', drift_args = drift_args, tolerance = X.shape[0]*0.05)

    def eval_fun(Y_obs, Y_hat):
        return sum((Y_obs - Y_hat)**2)

    test = DriftPredictionsResistanceTest(model	  = model,
                               X		  = X,
                               drift_type = 'scale_drift',
                               drift_args = drift_args,
                               eval		  = eval_fun,
                               tolerance  = X.shape[0]*0.05)

    test.run()

    schema = DataSchema().generate_manual(dataframe		   = X,
                                          categ_columns    = ['sex', 'education', 'marriage'],
                                          discrete_columns = [],
                                          binary_columns   = [])

    test = DriftPredictionsResistanceTest(model          = model,
                               X              = X,
                               drift_type     = 'scale_drift',
                               drift_args     = drift_args,
                               dataset_schema = schema,
                               tolerance      = X.shape[0]*0.05)

    test.run()

def test_drift_prediction_resistance_test_hard_labels_multiclass():
    X, y = make_classification(n_samples=300, n_features=3, n_informative=3, n_redundant=0, n_classes=3)
    X = pd.DataFrame(X, columns=["f"+str(i) for i in range(X.shape[1])])

    model = RandomForestClassifier()
    model = model.fit(X, y)

    drift_args = {'cols' : ['f0', 'f1', 'f2'], 'iqr' : [1, 1.1]}
    test = DriftPredictionsResistanceTest(model = model, X = X, drift_type = 'scale_drift', drift_args = drift_args, tolerance = X.shape[0]*0.5)
    test.run()
    assert test.info()["loss"] < test.tolerance

def test_drift_resistance_test():
    X, y = make_classification(n_samples=300, n_features=3, n_informative=3, n_redundant=0, n_classes=3)
    X = pd.DataFrame(X, columns=["f"+str(i) for i in range(X.shape[1])])

    model = RandomForestClassifier()
    model = model.fit(X, y)

    drift_args = {'cols' : ['f0', 'f1', 'f2'], 'iqr' : [1, 1.1]}
    test = DriftPredictionsResistanceTest(model = model, X = X, drift_type = 'scale_drift', drift_args = drift_args, tolerance = X.shape[0]*0.5)
    test.run()
    assert test.info()["loss"] < test.tolerance


CLASS_NEGATIVE = 0
CLASS_NEUTRAL = 1
CLASS_POSITIVE = 2
class SimpleModel():

    def __init__(self):
        pass

    def predict(self, sentences):
        """predicts the class with the mayority of words"""
        preds = []
        for sentence in sentences:
            if sentence.count("pos_word") > sentence.count("neg_word"):
                preds.append(CLASS_POSITIVE)
            elif sentence.count("neg_word") > sentence.count("pos_word"):
                preds.append(CLASS_NEGATIVE)
            else:
                preds.append(CLASS_NEUTRAL)
        return preds

def test_invariance_test_basic():

    # Create simple sample
    original = [
        "neutral pos_word",
        "neutral neg_word",
        "neutral neutral"
    ]

    perturbed = [
        ["neutral pos_word", "neutral pos_word"],
        ["neutral neutral", "neutral pos_word"],
        ["neutral pos_word", "irrelevant neutral"]
    ]

    model = SimpleModel()

    # Low threshold raises error
    test = ClassificationInvarianceTest(
        original_samples=original,
        perturbed_samples=perturbed,
        model=model,
        threshold=0.,
        check_total_errors_rate=True,
        name="test name"
    )
    assert test.info() is None

    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert_array_equal(test.pred_is_different[0] , np.array([False, False]))
    assert_array_equal(test.pred_is_different[1] , np.array([True, True]))
    assert_array_equal(test.pred_is_different[2] , np.array([True, False]))
    assert test.num_failed_per_sample == [0,2,1]
    assert test.num_perturbed_per_sample == [2,2,2]
    assert_array_equal(test.samples_with_errors, np.array([False , True, True]))
    assert round(test.rate_samples_with_errors,2) == 0.67
    assert round(test.info()["rate_samples_with_errors"],2) == 0.67
    assert len(test.get_examples_failed(n_samples = 2, n_perturbed = 1)) == 2

    # High threhsold doesn't raise error (Let's pass the predict_fn this time)
    test = ClassificationInvarianceTest(
        original_samples=original,
        perturbed_samples=perturbed,
        predict_fn=model.predict,
        threshold=0.9,
        check_total_errors_rate=False,
        name="test name"
    )
    test.run()
    assert_array_equal(test.pred_is_different[0] , np.array([False, False]))
    assert_array_equal(test.pred_is_different[1] , np.array([True, True]))
    assert_array_equal(test.pred_is_different[2] , np.array([True, False]))
    assert test.num_failed_per_sample == [0,2,1]
    assert test.num_perturbed_per_sample == [2,2,2]
    assert_array_equal(test.samples_with_errors, np.array([False , True, True]))
    assert round(test.rate_samples_with_errors,2) == 0.67


def test_invariance_test_bad_arguments():
    original = ["sample1", "sample2", "sample3"]
    perturbed = ["sample1_pertubed"] # Only contains 1 of the samples
    test = ClassificationInvarianceTest(
        original_samples=original,
        perturbed_samples=perturbed,
        predict_fn=lambda x: 1,
        threshold=0.,
        check_total_errors_rate=True,
        name="test name"
    )

    with pytest.raises(ValueError) as exinfo:
        test.run()

    # Neither predict function nor model are specified
    original = ["sample1"]
    perturbed = ["sample1_pertubed"] # Only contains 1 of the samples
    test = ClassificationInvarianceTest(
        original_samples=original,
        perturbed_samples=perturbed,
        threshold=0.,
        check_total_errors_rate=True,
        name="test name"
    )

    with pytest.raises(ValueError) as exinfo:
        test.run()

def test_tree_coverage_ok():
    x, y = make_classification(200)

    # Test ensemble sklearn
    rf = RandomForestClassifier().fit(x, y)
    test = TreeCoverageTest(rf, x)
    test.run()

    # Test decision tree
    t = DecisionTreeClassifier().fit(x, y)
    test = TreeCoverageTest(t, x)
    test.run()
    # nothing should have failed 'till this point

def test_tree_coverage_fail():
    x, y = make_regression(200)
    rf = DecisionTreeRegressor().fit(x, y)
    test = TreeCoverageTest(rf, np.zeros_like(x))

    with pytest.raises(FailedTestError):
        test.run()

    t = DecisionTreeRegressor().fit(x, y)
    test = TreeCoverageTest(t, np.zeros_like(x))

    with pytest.raises(FailedTestError):
        test.run()

def test_tree_coverage_fail():
    x, y = make_regression(200)
    rf = DecisionTreeRegressor().fit(x, y)
    test = TreeCoverageTest(rf, np.zeros_like(x))

    with pytest.raises(FailedTestError):
        test.run()

    t = DecisionTreeRegressor().fit(x, y)
    test = TreeCoverageTest(t, np.zeros_like(x))

    with pytest.raises(FailedTestError):
        test.run()

def test_tree_coverage_pipeline():
    x, y = make_classification(200)

    pipeline = Pipeline(steps=[
        ("scaler", StandardScaler()),
        ("classifier", RandomForestClassifier())
    ])
    pipeline = pipeline.fit(x, y)
    test = TreeCoverageTest(pipeline, x)
    test.run()

def test_tree_coverage_pipeline_without_tree_fails():
    x, y = make_classification(200)

    pipeline = Pipeline(steps=[
        ("scaler", StandardScaler()),
        ("classifier", LogisticRegression())
    ])
    pipeline = pipeline.fit(x, y)
    with pytest.raises(ValueError):
        test = TreeCoverageTest(pipeline, x)
        test.run()



def test_drift_metric_resistance_test_regression():

    X, y = make_regression(n_samples=200, n_features=3, n_informative=3)
    X = pd.DataFrame(X, columns=["f"+str(i) for i in range(X.shape[1])])

    model_lr = LinearRegression().fit(X, y)
    model_rf = RandomForestRegressor().fit(X, y)

    drift_args = {
        'cols' : ['f0'], 
        'method' : 'value',
        'method_params': {'value': 99999}
    }

    # Linear Regression is highly affected by extreme values
    test_lr = DriftMetricResistanceTest(
        model = model_lr,
        X = X,
        Y = y,
        drift_type = "outliers_drift",
        drift_args = drift_args,
        tolerance= 500,
    )
    with pytest.raises(FailedTestError):
        test_lr.run()

    # Random Forest is less afected
    test_rf = DriftMetricResistanceTest(
        model = model_rf,
        X = X,
        drift_type = "outliers_drift",
        drift_args = drift_args,
        tolerance= 500,
        Y = y
    )
    test_rf.run()

    assert test_lr.info()['metric_diff'] > test_rf.info()['metric_diff']

def test_drift_metric_resistance_test_classification():

    X, y = make_classification(n_samples=300, n_features=3, n_informative=3, n_redundant=0)
    X = pd.DataFrame(X, columns=["f"+str(i) for i in range(X.shape[1])])

    model = LogisticRegression().fit(X, y)
    
    drift_args_1 = {
        'cols' : ['f0', 'f1', 'f2'], 
        'method' : 'value',
        'method_params': {'value': 1e9, 'proportion_outliers': 0.99}
    }

    test_1 = DriftMetricResistanceTest(
        model = model,
        X = X,
        drift_type = "outliers_drift",
        drift_args = drift_args_1,
        tolerance= 1.0,
        Y = y
    )
    test_1.run()
    assert test_1.info()['metric_diff'] <= 1.

    # Test custom eval metric
    def custom_eval_metric(y_true, y_hat):
        return 5.

    test = DriftMetricResistanceTest(
        model = model,
        X = X,
        drift_type = "outliers_drift",
        drift_args = drift_args_1,
        tolerance= 10.0,
        Y = y,
        eval=custom_eval_metric
    )
    test.run()
    assert test.info()['metric_diff'] == 0.

def test_drift_metric_resistance_test_invalid_inputs():

    X, y = make_classification(n_samples=300, n_features=3, n_informative=3, n_redundant=0)
    X = pd.DataFrame(X, columns=["f"+str(i) for i in range(X.shape[1])])

    model = LogisticRegression().fit(X, y)
    
    drift_args_1 = {
        'cols' : ['f0', 'f1', 'f2'], 
        'method' : 'value',
        'method_params': {'value': 1e9, 'proportion_outliers': 0.99}
    }

    # invalid task
    test = DriftMetricResistanceTest(
        model = model,
        X = X,
        drift_type = "outliers_drift",
        drift_args = drift_args_1,
        tolerance= 1.0,
        Y = y,
        task='invalid'
    )
    with pytest.raises(ValueError):
        test.run()

def test_feature_checker_test():
    df = pd.read_csv('%s/csv/test_twt.csv' % pathlib.Path(__file__).parent.resolve(), sep = '\t')

    train, test = train_test_split(df, test_size = 0.2)

    X_train = train.loc[:, train.columns != 'Y']
    Y_train = train.loc[:, 'Y']

    rf = RandomForestClassifier()
    rf.fit(X_train, Y_train)

    fct = FeatureCheckerTest(rf, train, 'Y', tolerance = 100)
    with pytest.raises(FailedTestError):
        fct.run()

    fct = FeatureCheckerTest(rf, train, 'Y', test, tolerance = 100)
    with pytest.raises(FailedTestError):
        fct.run()

    def eval_fun(Y_obs, Y_hat):
        return sum((Y_obs - Y_hat)**2)

    fct = FeatureCheckerTest(rf, train, 'Y', test, eval = eval_fun, tolerance = 100)
    with pytest.raises(FailedTestError):
        fct.run()

    fct = FeatureCheckerTest(rf, train, 'Y', test, importance = 'feature_importances_', tolerance = 100)
    with pytest.raises(FailedTestError):
        fct.run()

    fct = FeatureCheckerTest(rf, train, 'Y', test, importance = 'feature_importances_', eval = eval_fun, tolerance = 100, remove_num = 2, num_tries = 1)
    with pytest.raises(FailedTestError):
        fct.run()


def test_feature_checker_test_model_function():

    # Create Dataset
    X, y = make_classification(200, n_features=5, n_informative=5, n_redundant=0)
    df = pd.DataFrame(X, columns=["x1", "x2", "x3", "x4", "x5"])
    df["x6"] = np.random.choice(["cat_0", "cat_1", "cat_2"], size=len(df)) # make categorical feature feature
    df["y"] = y
    df_train, df_test = train_test_split(df, test_size = 0.2)

    # Define function to create the model (pipeline)
    def create_pipeline_fn(X, model_args):

        num_feats = [f for f in model_args['num_feats'] if f in X.columns]
        cat_feats = [f for f in model_args['cat_feats'] if f in X.columns]

        numeric_transformer = Pipeline(steps=[("scaler", StandardScaler())])
        categorical_transformer = Pipeline(steps=[("ohe", OneHotEncoder(handle_unknown="ignore"))])

        preprocessor = ColumnTransformer(
            transformers=[
                ("num", numeric_transformer, num_feats),
                ("cat", categorical_transformer, cat_feats),
            ], remainder='drop'
        )

        pipeline = Pipeline(
            steps=[
                ("preprocessor", preprocessor),
                ("classifier", DecisionTreeClassifier())
            ]
        )
        
        return pipeline

    # Create and run test
    num_feats = ["x1", "x2", "x3", "x4", "x5"]
    cat_feats = ["x6"]
    model_fn_args = {
        'num_feats': num_feats,
        'cat_feats': cat_feats
    }
    test = FeatureCheckerTest(
        model=create_pipeline_fn,
        train=df_train,
        target="y",
        test=df_test,
        tolerance=-1,
        num_tries=3,
        remove_num=1,
        model_fn_args = model_fn_args,
    )
    test.run()


import pytest
import seaborn as sns
import numpy as np
import pandas as pd

from mercury.dataschema import DataSchema
from mercury.dataschema.feature import FeatType
from mercury.robust.data_tests import (
    SameSchemaTest,
    DriftTest,
    LinearCombinationsTest,
    LabelLeakingTest,
    NoisyLabelsTest,
    CohortPerformanceTest,
    SampleLeakingTest,
    NoDuplicatesTest
)
from mercury.robust.errors import FailedTestError

from sklearn import preprocessing
from sklearn.metrics import accuracy_score
from sklearn.compose import ColumnTransformer
from sklearn.pipeline import Pipeline
from sklearn.linear_model import LogisticRegression
from sklearn.datasets import make_classification


@pytest.fixture(scope="module")
def datasets():
    tips = sns.load_dataset("tips")
    tips["sex"] = tips["sex"].astype(str)
    tips["smoker"] = tips["smoker"].astype(str)
    tips["day"] = tips["day"].astype(str)
    tips["time"] = tips["time"].astype(str)

    titanic = sns.load_dataset("titanic")
    isna_deck = titanic.deck.isna()
    titanic["class"] = titanic["class"].astype(str)
    titanic["deck"] = titanic["deck"].astype(str)
    titanic["who"] = titanic["who"].astype(str)
    titanic.loc[isna_deck, "deck"] = np.nan

    return tips, titanic


def test_sameschema(datasets):
    tips, titanic = datasets
    titanic2 = titanic.copy()
    titanic2 = titanic.drop("deck", axis=1)

    schma = DataSchema().generate(titanic).calculate_statistics()
    schma2 = DataSchema().generate(titanic2).calculate_statistics()

    with pytest.raises(FailedTestError) as exinfo:
        SameSchemaTest(titanic2, schma).run()

    # SameSchemaTest(titanic, schma, schema_cat_threshold=0.01).run()
    SameSchemaTest(titanic, schma).run()

def test_sameschema_custom_feature_map(datasets):

    tips, titanic = datasets
    titanic2 = titanic.copy()

    force_types = {
        'sex': FeatType.CATEGORICAL,
        'pclass': FeatType.DISCRETE
    }
    schma_reference = DataSchema().generate(titanic, force_types=force_types).calculate_statistics()

    # Same types are forced. the test runs fine
    test = SameSchemaTest(
        titanic2, 
        schma_reference,
        custom_feature_map={
            'pclass': FeatType.DISCRETE,
            'sex': FeatType.CATEGORICAL
        }
    )
    test.run()


def test_drift(datasets):
    tips, titanic = datasets

    schma = DataSchema().generate(titanic).calculate_statistics()

    # Simulate a dataset with drift on the age feature
    titanic2 = titanic.copy()
    titanic2["age"] *= titanic2["age"]  # quadratic relation

    drift_test = DriftTest(titanic2, schma)
    assert drift_test.info() is None
    with pytest.raises(FailedTestError) as exinfo:
        drift_test.run()

    assert "age" in str(exinfo.value)

    # Simulate a dataset with drift on the sex feature. Only males are present
    titanic2 = titanic.copy()
    titanic2["sex"] = "male"

    with pytest.raises(FailedTestError) as exinfo:
        DriftTest(titanic2, schma).run()

    assert "sex" in str(exinfo.value)

    # This shouldn't fail
    drift_test = DriftTest(titanic, schma)
    drift_test.run()
    assert 'cat_drift_metrics' in drift_test.info()

    # Dataset without any Continuous Features should work without raising exception (Previous bug)
    titanic_cat = titanic[["sex", "embarked"]]
    schma_cat = DataSchema().generate(titanic_cat).calculate_statistics()
    drift_test = DriftTest(titanic_cat, schma_cat)
    drift_test.run()
    # Dataset without any Categorical Feature should work without raising exception (Previous bug)
    titanic_cont = titanic[["age", "fare"]]
    schma_cont = DataSchema().generate(titanic_cont).calculate_statistics()
    drift_test = DriftTest(titanic_cont, schma_cont)
    drift_test.run()

def test_drift_different_domain_categoricals(datasets):

    tips, titanic = datasets

    # Same number of categories, but one category changes in the test set
    titanic_ref = titanic[["sex", "embarked"]].copy()
    titanic_test = titanic_ref.copy()

    titanic_test = titanic_test.replace({'sex': {'male': 'm'}})

    schma_ref = DataSchema().generate(titanic_ref).calculate_statistics()
    drift_test = DriftTest(titanic_test, schma_ref)
    with pytest.raises(FailedTestError) as exinfo:
        drift_test.run()


    # test set has one category more
    titanic_ref = titanic[["sex", "embarked"]].copy()
    titanic_test = titanic_ref.copy()

    new_records = titanic_test.iloc[0:5].copy()
    new_records['embarked'] = 'new_value'
    titanic_test = pd.concat([titanic_test, new_records])

    schma_ref = DataSchema().generate(titanic_ref).calculate_statistics()
    drift_test = DriftTest(titanic_test, schma_ref)
    with pytest.raises(FailedTestError) as exinfo:
        drift_test.run()


    # Test has one category less, in this case it fails because of drift
    titanic_ref = titanic[["sex", "embarked"]].copy()
    titanic_test = titanic_ref.copy()
    titanic_test = titanic_test[titanic_test['embarked'] != 'Q']

    schma_ref = DataSchema().generate(titanic_ref).calculate_statistics()
    drift_test = DriftTest(titanic_test, schma_ref)
    with pytest.raises(FailedTestError) as exinfo:
        drift_test.run()

def test_drift_target_samples_outside_source_bin_limits():
    # This is is testing the case where the target distribution in a continous variable contains samples 
    # that are outside of the limits of the binning created in the source distribution


    # Case 1: Target samples contain values higher than the max bin
    df_source = pd.DataFrame(
        data=[0, 0.5, 1, 2., 2.3, 5, 6., 7, 8, 9, 9.5, 10],
        columns=["f1"]
    )
    df_target = pd.DataFrame(
        data=[11., 11.3, 12, 13, 14, 15, 16, 17],
        columns=["f1"]
    )
    schema_source = DataSchema().generate(df_source).calculate_statistics()
    drift = DriftTest(df_target, schema_ref=schema_source)
    with pytest.raises(FailedTestError) as exinfo:
        drift.run()

    # Case 2: Target samples contain values lower than min bin and higher than max bin
    df_source = pd.DataFrame(
        data=[0, 0.5, 1, 2., 2.3, 5, 6., 7, 8, 9, 9.5, 10],
        columns=["f1"]
    )
    df_target = pd.DataFrame(
        data=[-4., -3., -1, 3., 5.3, 7, 9, 14, 15, 16, 17],
        columns=["f1"]
    )
    schema_source = DataSchema().generate(df_source).calculate_statistics()
    drift = DriftTest(df_target, schema_ref=schema_source)
    with pytest.raises(FailedTestError) as exinfo:
        drift.run()

    # Case 3: Only one sample (of many) outside de limits of distribution shouldn't produce drift
    df_source = pd.DataFrame(
        data=[0.]*20 + [1.1]*20 + [3.3]*20 + [6.6]*20 + [7.7]*20 + [9.9]*20 + [10.]*20 + [12.]*20 + [15.]*20 + [20.]*20,
        columns=["f1"]
    )
    df_target = pd.DataFrame(
        data=[0.]*20 + [1.1]*20 + [3.3]*20 + [6.6]*20 + [7.7]*20 + [9.9]*20 + [10.]*20 + [12.]*20 + [15.]*20 + [20.]*20 + [23.],
        columns=["f1"]
    )
    schema_source = DataSchema().generate(df_source).calculate_statistics()
    drift = DriftTest(df_target, schema_ref=schema_source)
    drift.run()


def test_label_leaking(datasets):
    tips, titanic = datasets

    titanic = titanic.dropna()
    tips = tips.dropna()

    # Remove category
    tips["sex"] = tips["sex"].astype(str)
    tips["smoker"] = tips["smoker"].astype(str)
    tips["day"] = tips["day"].astype(str)
    tips["time"] = tips["time"].astype(str)

    # Label encoding
    for feat in ["sex", "smoker", "day", "time"]:
        tips.loc[:, feat] = preprocessing.LabelEncoder().fit_transform(
            tips.loc[:, feat]
        )

    # Categoricals need to be label encoded
    le = preprocessing.LabelEncoder()
    categoricals = [
        "sex",
        "class",
        "who",
        "deck",
        "embark_town",
        "alive",
        "alone",
        "embarked",
    ]
    for f in categoricals:
        titanic[f] = le.fit_transform(titanic.loc[:, f])

    # Only classification / regression possible
    with pytest.raises(ValueError) as exinfo:
        test = LabelLeakingTest(titanic, task="wrong", label_name="survived")
        test.run()
    assert "classification" in str(exinfo.value)

    test = LabelLeakingTest(titanic, label_name="survived")

    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert "alive" in str(exinfo.value)

    # This will not fail
    titanic2 = titanic.drop("alive", axis=1)

    x, y = make_classification(n_samples=10_000, n_features=30)
    x = pd.DataFrame(x)
    x["target"] = y
    x["injected"] = y + np.random.normal(0, 0.2, len(y))

    test = LabelLeakingTest(
        x,
        label_name="target",
    )

    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert "injected" in str(exinfo.value)

    # test with regression and inject a nonlinear relationship
    tips["injected"] = tips["total_bill"] ** 2
    test = LabelLeakingTest(tips, label_name="total_bill", threshold=0.03)

    with pytest.raises(FailedTestError) as exinfo:
        test.run()
    assert "injected" in str(exinfo.value)

    # Test ignoring features.
    # In this case, it will pass because we ignore the injected features
    test = LabelLeakingTest(
        tips, label_name="total_bill", threshold=0.03, ignore_feats=["injected"]
    )
    test.run()


def test_linear_combs_continuous():
    custom_feature_mapping = {
        0: FeatType.CONTINUOUS,
        1: FeatType.CONTINUOUS,
        2: FeatType.CONTINUOUS
    }

    X = np.array([[1.0, 1.0, 1.0], [2.0, 1.0, 2.0], [3.0, 2.0, 3.0], [4.0, 3.0, 4.0]])

    with pytest.raises(FailedTestError) as exinfo:
        LinearCombinationsTest(pd.DataFrame(X), schema_custom_feature_map=custom_feature_mapping).run()
    assert "Linear combinations for continuous features were encountered" in str(exinfo.value)

    X = np.array([[-1.0, 4.0, 2.0], [4.0, 6.0, 3.0], [4.0, 8.0, 4.0], [3.0, 10.0, 5.0]])

    with pytest.raises(FailedTestError) as exinfo:
        LinearCombinationsTest(pd.DataFrame(X), schema_custom_feature_map=custom_feature_mapping).run()
    assert "Linear combinations for continuous features were encountered" in str(exinfo.value)

    # This should pass
    X = np.array([[0.0, 1.0, 2.0], [2.0, 2.0, 5.0], [5.0, 1.0, 7.0], [0.0, 9.0, 2.0]])
    LinearCombinationsTest(pd.DataFrame(X), schema_custom_feature_map=custom_feature_mapping).run()


def test_linear_combs_categoricals():
    cps = df = pd.DataFrame({
        "cp": [
            "28701",
            "28015",
            "28710",
            "48015",
            "48001",
            "01001",
            "01002",
            "01003",
            "28701",
        ],
        "prov": [
            "Madrid",
            "Madrid",
            "Madrid",
            "Vizcaya",
            "Vizcaya",
            "Álava",
            "Álava",
            "Álava",
            "Madrid",
        ],
        "num": [5432.1, 243.3, 17.6, 22.4, 97.4, 1.2, -2.2, 56.7, 246.0],
    })

    with pytest.raises(FailedTestError) as exinfo:
        LinearCombinationsTest(cps).run()

    assert "prov" in str(exinfo.value)

def test_noisy_labels_tabular(datasets):
    tips, titanic = datasets
    titanic = titanic[["survived", "pclass", "sex", "age", "fare", "who"]].dropna().copy()

    # this test should pass
    from sklearn.ensemble import RandomForestClassifier
    label_issues_args = dict(
        clf=RandomForestClassifier(max_depth=4)
    )
    test = NoisyLabelsTest(
        titanic,
        label_name='survived',
        threshold=0.25,
        label_issues_args=label_issues_args
    )
    test.run()

    # It should pass also using calculate_idx_issues=True
    test = NoisyLabelsTest(
        titanic,
        label_name='survived',
        threshold=0.25,
        calculate_idx_issues=True,
        label_issues_args=label_issues_args
    )
    test.run()

    # this test shouldn't pass
    titanic2 = titanic.copy()
    titanic2["survived"] = np.random.choice([0, 1], replace=True, size=len(titanic2))
    test = NoisyLabelsTest(
        titanic2,
        label_name='survived',
        threshold=0.25,
        label_issues_args=label_issues_args
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()

    # Case using own preprocessor
    preprocessor = ColumnTransformer(
        transformers=[('ohe', preprocessing.OneHotEncoder(handle_unknown='ignore'), ["sex", "who"])],
        remainder='passthrough'
    )
    test = NoisyLabelsTest(
        titanic,
        label_name='survived',
        threshold=0.25,
        preprocessor=preprocessor,
        label_issues_args=label_issues_args
    )
    test.run()

    # Case prprocessingis not needed (only continous features)
    test = NoisyLabelsTest(
        titanic[["age", "fare", "survived"]],
        label_name='survived',
        threshold=0.35,
        label_issues_args=label_issues_args
    )
    test.run()

    # Case using ignore features and different configuration for cleanlab
    label_issues_args = dict(
        clf=RandomForestClassifier(max_depth=4),
        sorted_index_method='normalized_margin'
    )
    test = NoisyLabelsTest(
        titanic,
        label_name='survived',
        ignore_feats=["pclass"],
        threshold=0.25,
        calculate_idx_issues=True,
        label_issues_args=label_issues_args
    )
    test.run()


def test_noisy_labels_text():

    df = pd.DataFrame({
        "text": [
            "positive text",
            "positive text",
            "positive text",
            "positive text",
            "positive text",
            "negative text",
            "negative text",
            "negative text",
            "negative text",
            "negative text"
        ],
        "label": [1, 1, 1, 1, 1, 0, 0, 0, 0, 0],
    })

    # this test should pass
    test = NoisyLabelsTest(
        df,
        label_name='label',
        text_col='text',
        threshold=0.25,
    )
    test.run()

    # this test should fail
    size = 500
    df = pd.DataFrame()
    df["text"] = ["text class0"] * 25 + ["text class1"] * 25 + ["text class2"] * 25 + ["text class3"] * 25
    df["label"] = [0] * 25 + [1] * 25 + [2] * 25 + [3] * 25
    df["label"] = np.random.choice([0,1, 2, 3], size=len(df), replace=True)
    test = NoisyLabelsTest(
        df,
        label_name='label',
        text_col='text',
        threshold=0.25,
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()

    # Case creating own text feature extraction
    from sklearn.feature_extraction.text import CountVectorizer
    cv = CountVectorizer(max_features=1000)
    test = NoisyLabelsTest(
        df,
        label_name='label',
        text_col='text',
        threshold=0.20,
        preprocessor=cv
    )
    with pytest.raises(FailedTestError) as exinfo:
        test.run()


def test_cohort_performance(datasets):
    tips, titanic = datasets
    titanic = titanic[["survived", "age", "fare", "pclass", "sex"]].dropna()

    label_col = "survived"

    # Create pipeline to train model
    numeric_features = ['age', 'fare']
    numeric_transformer = Pipeline(steps=[("scaler", preprocessing.StandardScaler())])
    categorical_features = ['pclass', 'sex']
    categorical_transformer = preprocessing.OneHotEncoder(handle_unknown="ignore")
    preprocessor = ColumnTransformer(
        transformers=[
            ("num", numeric_transformer, numeric_features),
            ("cat", categorical_transformer, categorical_features),
        ]
    )
    pipeline = Pipeline(
        steps=[("preprocessor", preprocessor), ("classifier", LogisticRegression(solver='newton-cg'))]
    )
    # Train
    X_train = titanic[numeric_features + categorical_features]
    y_train = titanic[label_col]
    pipeline = pipeline.fit(X_train, y_train)
    # Generate Predictions
    titanic2 = titanic.copy()
    titanic2["class_pred"] = pipeline.predict(X_train)
    probs = pipeline.predict_proba(X_train)
    col_probs = []
    for i in range(probs.shape[1]):
        titanic2["prob_" + str(i)] = probs[:,i]

    # Test 1 - Accuracy (compare_max_diff=True) - SHOULD PASS
    def eval_acc(df):
        return accuracy_score(df[label_col], df["class_pred"])
    test1 = CohortPerformanceTest(
        base_dataset=titanic2, group_col="sex", eval_fn = eval_acc, threshold = 0.2, compare_max_diff=True,
        threshold_is_percentage = False
    )
    assert test1.info() == None
    test1.run()
    assert test1.metric_by_group.max() - test1.metric_by_group.min() < test1.threshold
    assert 'metric_by_group' in test1.info()

    # Test 2 - Accuracy (compare_max_diff=False) - SHOULD PASS
    test2 = CohortPerformanceTest(
        base_dataset=titanic2, group_col="sex", eval_fn = eval_acc, threshold=0.2, compare_max_diff=False,
        threshold_is_percentage=False
    )
    test2.run()
    assert test2.mean_metric - test1.metric_by_group.min() < test2.threshold

    # Test 3 - Mean Prediction (compare_max_diff=TRUE) SHOULD NOT PASS
    def eval_mean_prediction(df):
        return df["prob_1"].mean()

    test3 = CohortPerformanceTest(
        base_dataset=titanic2, group_col="sex", eval_fn = eval_mean_prediction, threshold = 0.1,
        compare_max_diff=True, threshold_is_percentage=False
    )
    with pytest.raises(FailedTestError) as exinfo:
        test3.run()

    # Test 4 - Mean Prediction (compare_max_diff=FALSE) SHOULD NOT PASS
    test4 = CohortPerformanceTest(
        base_dataset=titanic2, group_col="sex", eval_fn = eval_mean_prediction, threshold = 0.1,
        compare_max_diff=False, threshold_is_percentage=False
    )
    with pytest.raises(FailedTestError) as exinfo:
        test4.run()

    # Test 5 - Mean Prediction (compare_max_diff=True, threshold_is_percentage=True) - SHOULD FAIL
    test5 = CohortPerformanceTest(
        base_dataset=titanic2, group_col="sex", eval_fn = eval_mean_prediction, threshold = 1., compare_max_diff=True,
        threshold_is_percentage = True
    )
    with pytest.raises(FailedTestError) as exinfo:
        test5.run()

    # Test 6 - Mean Predictin (compare_max_diff=False, threshold_is_percentage=True) - SHOULD PASS
    test6 = CohortPerformanceTest(
        base_dataset=titanic2, group_col="sex", eval_fn = eval_mean_prediction, threshold = 1.5, compare_max_diff=False,
        threshold_is_percentage = True
    )
    test6.run()

def test_label_leaking_test_str(datasets):
    tips, titanic = datasets
    titanic = titanic.dropna()

    # Raises ValueError becuase handle_str_cols indicates to raise Error when finding string columns
    test = LabelLeakingTest(titanic, task="classification", label_name="survived", handle_str_cols='error')
    with pytest.raises(ValueError) as exinfo:
        test.run()

    # String columns are transformed the test is executed. Raises FailedTestError because leaking in alive column
    test = LabelLeakingTest(titanic, task="classification", label_name="survived", handle_str_cols='transform')
    with pytest.raises(FailedTestError) as exinfo:
        test.run()

    # Raises ValueError because hand_src_cols param has a wrong value
    test = LabelLeakingTest(titanic, task="classification", label_name="survived", handle_str_cols='wrong value')
    with pytest.raises(ValueError) as exinfo:
        test.run()

def test_sample_leaking_test_numbers():

    synthetic_train = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [0.7, 0.8, 0.9],
        [1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5],
        [1.6, 1.7, 1.8],
    ], columns=["f1", "f2", "f3"])


    # Contains 2 repeated
    synthetic_test = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [3.7, 3.8, 3.9],
        [4.0, 4.1, 4.2]
    ], columns=["f1", "f2", "f3"])

    # Allowing 1 duplicate the test fails
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=1)
    with pytest.raises(FailedTestError):
        test.run()
    assert test.info()['num_duplicated'] == 2
    # Check for Hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=1, use_hash=True)
    with pytest.raises(FailedTestError):
        test.run()
    assert test.info()['num_duplicated'] == 2

    # Allowing 3 duplicates the test passes
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=3)
    test.run()
    assert test.info()['num_duplicated'] == 2
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=3, use_hash=True)
    test.run()
    assert test.info()['num_duplicated'] == 2

    # Allowing a percentage of 0.25 duplicated samples the test fails
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0.25, use_hash=False)
    with pytest.raises(FailedTestError):
        test.run()
    assert test.info()['percentage_duplicated'] >= 0.1
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0.25, use_hash=True)
    with pytest.raises(FailedTestError):
        test.run()
    assert test.info()['percentage_duplicated'] >= 0.1

    # Allowing a percentage of 0.75 duplicated samples the test fails
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0.75)
    test.run()
    assert test.info()['percentage_duplicated'] <= 0.75
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0.75, use_hash=True)
    test.run()
    assert test.info()['percentage_duplicated'] <= 0.75


def test_sample_leaking_test_ignore_feats_param():

    synthetic_train = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [0.7, 0.8, 0.9],
        [1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5],
        [1.6, 1.7, 1.8],
    ], columns=["f1", "f2", "f3"])

    # Doesn't contain repeated thanks to f3. If f3 removed all are repeated with synthetic_train
    synthetic_test_1 = pd.DataFrame(data=[
        [0.1, 0.2, 1.3],
        [0.4, 0.5, 1.6],
        [0.7, 0.8, 1.9],
        [1.0, 1.1, 2.2]
    ], columns=["f1", "f2", "f3"])

    # Without ignoring any feature the test passes
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test_1, ignore_feats=None, threshold=0)
    test.run()

    # Ignoring column 3 raises error
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test_1, ignore_feats=['f3'], threshold=0)
    with pytest.raises(FailedTestError):
        test.run()

def test_sample_leaking_test_strings():

    synthetic_train = pd.DataFrame(data=[
        [0.1, "val_1", "value_1"],
        [0.4, "val_2", "value_1"],
        [0.7, "val_3", "value_1"],
        [1.0, "val_4", "value_1"]
    ], columns=["f1", "f2", "f3"]
    )

    # Doesn't contain repeated thanks to f3
    synthetic_test = pd.DataFrame(data=[
        [0.1, "val_1", "value_2"],
        [0.4, "val_2", "value_2"],
        [0.7, "val_3", "value_2"],
        [1.0, "val_4", "value_2"]
    ], columns=["f1", "f2", "f3"]
    )

    # Using all columns the test passes
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None)
    test.run()
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, use_hash=True)
    test.run()

    # Ignoring column f3 raises error
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=['f3'])
    with pytest.raises(FailedTestError):
        test.run()
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=['f3'], use_hash=True)
    with pytest.raises(FailedTestError):
        test.run()

def test_sample_leaking_test_contains_duplicated_only_in_train():

    synthetic_train = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.1, 0.2, 0.3],
        [0.7, 0.8, 0.9],
        [1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5],
        [1.6, 1.7, 1.8],
    ], columns=["f1", "f2", "f3"])

    synthetic_test = pd.DataFrame(data=[
        [10.1, 10.2, 10.3],
        [10.4, 10.5, 10.6],
        [13.7, 13.8, 13.9],
        [14.0, 14.1, 14.2]
    ], columns=["f1", "f2", "f3"])

    # test should pass
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0)
    test.run()
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0, use_hash=True)
    test.run()

def test_sample_leaking_test_contains_duplicated_only_in_test():

    synthetic_train = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [0.7, 0.8, 0.9],
        [1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5],
        [1.6, 1.7, 1.8],
    ], columns=["f1", "f2", "f3"])

    synthetic_test = pd.DataFrame(data=[
        [10.1, 10.2, 10.3],
        [10.1, 10.2, 10.3],
        [3.7, 3.8, 3.9],
        [4.0, 4.1, 4.2]
    ], columns=["f1", "f2", "f3"])

    # test should pass
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0)
    test.run()
    # Check for hash implementation
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None, threshold=0, use_hash=True)
    test.run()

def test_sample_leaking_test_different_columns_in_datasets():
    synthetic_test = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [0.7, 0.8, 0.9],
        [1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5],
        [1.6, 1.7, 1.8],
    ], columns=["f1", "f2", "f4"])


    # Contains 2 repeated
    synthetic_train = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, 0.6],
        [3.7, 3.8, 3.9],
        [4.0, 4.1, 4.2]
    ], columns=["f1", "f2", "f3"])

    # Allowing 1 duplicate the test fails
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None,
                            threshold=1)
    with pytest.raises(ValueError):
        test.run()

def test_dataset_with_nans():
    synthetic_train = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, np.nan],
        [0.7, 0.8, 0.9],
        [1.0, 1.1, 1.2],
        [1.3, 1.4, 1.5],
        [1.6, 1.7, 1.8],
    ], columns=["f1", "f2", "f3"])


    # Contains 2 repeated
    synthetic_test = pd.DataFrame(data=[
        [0.1, 0.2, 0.3],
        [0.4, 0.5, np.nan],
        [3.7, 3.8, 3.9],
        [4.0, 4.1, 4.2]
    ], columns=["f1", "f2", "f3"])

    # Allowing 1 duplicate the test fails
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None,
                            threshold=1)
    with pytest.raises(FailedTestError):
        test.run()
    assert test.info()['num_duplicated'] == 2

    # Allowing 3 duplicates the test passes
    test = SampleLeakingTest(base_dataset=synthetic_train, test_dataset=synthetic_test, ignore_feats=None,
                            threshold=3)
    test.run()
    assert test.info()['num_duplicated'] == 2

def test_no_duplicated_test():
    # Assert crashes with duplicated rows
    test_df = pd.DataFrame([
        [1, 2, 3],
        [2, 2, 4],
        [4, 4, 4],
        [4, 2, 3],
        [1, 2, 3],
        [1, 2, 3],
    ])

    test = NoDuplicatesTest(test_df)

    with pytest.raises(FailedTestError):
        test.run()

    # Assert all OK with no duplicates
    test_df = pd.DataFrame([
        [1, 2, 3],
        [2, 2, 4],
        [4, 4, 4],
        [4, 2, 3],
    ])

    test = NoDuplicatesTest(test_df)
    test.run()  # This should run with no problems

    # Force hash sample comparison
    test_df = pd.DataFrame([
        [1, 2, 3],
        [2, 2, 4],
        [4, 4, 4],
        [4, 2, 3],
        [1, 2, 3],
        [1, 2, 3],
    ])
    test = NoDuplicatesTest(test_df, use_hash=True)
    with pytest.raises(FailedTestError):
        test.run()

    # Check it works ignoring some features
    test = NoDuplicatesTest(test_df, ignore_feats=[2])
    with pytest.raises(FailedTestError):
        test.run()

def test_lin_combinations_cont():

    df = pd.DataFrame()
    df["f1"] = np.random.uniform(size=100)
    df["f2"] = df["f1"] * 2
    df["f3"] = np.random.uniform(size=100)
    df["f4"] = df["f1"] + df["f2"]
    df["f5"] = np.random.uniform(size=100)

    df["f6"] = np.random.uniform(size=100)
    df["f7"] = df["f6"] * 3
    df["f8"] = df["f6"] * 0.1

    schma_reference = DataSchema().generate(df).calculate_statistics()
    linear_comb_test = LinearCombinationsTest(df, dataset_schema=schma_reference)
    with pytest.raises(FailedTestError, match="'f1', 'f2'"):
        linear_comb_test.run()



import numpy as np


class CategoryStruct:
    """
    This is an abstract class containing some methods to identify structure among the categorical columns in a pandas dataframe.

    By structure we mean that the unique cases defined by a smaller subset of the categorical columns are the same as the unique
    cases defined by all the categorical columns, which implies that some columns could be removed.

    The simplest examples could be: The week day in German does not add any information when we already have it in English. The
    column set {DayEnglish, DayGerman} has the same information as {DayEnglish} (or as {DayGerman} obviously).
    A variable "province" does not add any information that is not already in "postal_code" (assuming postal codes are finer grained
    divisions that subdivide provinces.)

    The decision of what (if any) columns should be removed is not made by the methods. The methods just provide information to
    assist decision (see the individual docstrings).
    """

    @staticmethod
    def categoricals(data):
        """
        Returns a list of column names whose type is a Python object. Categoricals encoded as strings are Python objects in Pandas.
        In all the class methods, when for the argument `categoricals` is not given, this method will be called to define what the
        categorical variables are.

        Args:
            data (pandas.dataframe): The dataframe

        Returns:
            A list with the names of the columns whose type is a Python object
        """
        return [
            name
            for name, type in zip(data.columns, data.dtypes)
            if type == np.dtype("object")
        ]

    @staticmethod
    def cardinality_set_of_rows(data, col_names):
        """
        Returns the cardinality of the set of rows (for the given columns in col_names).

        This function is the workhorse of the whole CategoryStruct class. Everything is just about comparing the cardinalities of
        different sets of columns.

        Args:
            data (pandas.dataframe): The dataframe
            col_names(list of str):	 A list with the names of the columns to be selected

        Returns:
            The cardinality of the set of rows at the specified columns
        """
        if len(col_names) == 0:
            return 1

        return data.drop_duplicates(subset=col_names).shape[0]

    @staticmethod
    def synonyms(data, categoricals=None):
        """
        Searches for variable pairs that contain the same information.

        Args:
            data (pandas.dataframe):	The dataframe
            categoricals(list of str):	A list with the names of all the categorical columns to be searched

        Returns:
            A list of tuples of variable name pairs that contain the same information, typically encoded differently.
        """
        if categoricals is None:
            categoricals = CategoryStruct.categoricals(data)

        ret = []
        if len(categoricals) < 2:
            return ret

        card_name = []
        for name in categoricals:
            card_name.append(
                (CategoryStruct.cardinality_set_of_rows(data, [name]), name)
            )

        card_name.sort()

        for i, (c, name) in enumerate(card_name):
            for j in range(i + 1, len(card_name)):
                if card_name[j][0] != c:
                    break

                candidate = (name, card_name[j][1])
                if CategoryStruct.cardinality_set_of_rows(data, list(candidate)) == c:
                    ret.append(candidate)

        return ret

    @staticmethod
    def all_complete_sets(data, categoricals=None):
        """
        Returns all the complete sets of variables that have the same information as the given set.

        **Note**: This method is the slowest of all, but it already returns all the possible final choices of
        categorical variables that contain the same information as the given set.

        Args:
            data (pandas.dataframe):	The dataframe
            categoricals(list of str):	A list with the names of all the categorical columns to be searched

        Returns:
            A list of lists of the names of variables that are complete sets. This includes the original set passed to the function.
        """
        if categoricals is None:
            categoricals = CategoryStruct.categoricals(data)

        ret = [categoricals]
        if len(categoricals) < 2:
            return ret

        card_all_in = CategoryStruct.cardinality_set_of_rows(data, categoricals)

        for i, col in enumerate(categoricals):
            excluded = [col]
            remaining = [c for c in categoricals if c not in excluded]
            card_others = CategoryStruct.cardinality_set_of_rows(data, remaining)

            if card_others == card_all_in:
                ret.append(remaining)

                for j in range(i + 1, len(categoricals)):
                    excluded.append(categoricals[j])
                    remaining = [c for c in categoricals if c not in excluded]
                    card_others = CategoryStruct.cardinality_set_of_rows(
                        data, remaining
                    )

                    if card_others == card_all_in:
                        ret.append(remaining)
                    else:
                        excluded.pop()

        return ret

    @staticmethod
    def individually_redundant(data, categoricals=None):
        """
        Returns all the variables that could **individually** be removed.

        **Note that**: any of the variables returned can be removed individually. Meaning that only one of them should be removed
        at each step and this method should be called again on the resulting set of columns before removing another variable.
        Removing any variable may change the possible redundancy of the remaining variables.

        Args:
            data (pandas.dataframe):	The dataframe
            categoricals(list of str):	A list with the names of all the categorical columns to be searched

        Returns:
            A list with the names of all variables that are candidates to be removed, but only one at a time.
        """
        if categoricals is None:
            categoricals = CategoryStruct.categoricals(data)

        ret = []
        if len(categoricals) < 2:
            return ret

        card_all_in = CategoryStruct.cardinality_set_of_rows(data, categoricals)

        for col in categoricals:
            others = [c for c in categoricals if c != col]
            card_others = CategoryStruct.cardinality_set_of_rows(data, others)

            if card_others == card_all_in:
                ret.append(col)

        return ret


import numpy as np
from sklearn.linear_model import LogisticRegression
from cleanlab.count import estimate_cv_predicted_probabilities, compute_confident_joint
from cleanlab.filter import find_label_issues
from typing import Tuple


def get_confident_joint(
    X: np.array,
    y: np.array,
    psx: np.array = None,
    n_folds: int = 5,
    clf: "BaseEstimator" = None,  # noqa: F821
    seed: int = None
) -> np.array:

    """
    Calculates the confident joint matrix, which contains the counts of each one of the samples of a class that could belong to
    each one of the other classes. The diagonal entries of the matrix count correct labels and non-diagonal possible asymmetric
    label error counts. It uses the cleanlab library to compute it. More information in the Confident Learning paper:
    https://arxiv.org/pdf/1911.00068.pdf

    Args:
        X (np.array): array of features of your dataset
        y (np.array): array of (noisy) labels of your dataset, i.e. some labels may be erroneous.
        psx (np.array): matrix with (noisy) model-predicted probabilities for each of the N samples of X. It should be
            computed using cross-validation. If None, then it will be computed using `n_folds` and `clf`
        n_folds (int): number of cross-validation folds used to compute out-of-sample probabilities for each sample
        clf (BaseEstimator): Classifier to calculate out-of-sample probabilities. If None, LogisticRegression is used.
        seed (int): seed to use.

    Returns:
        Confident joint matrix

    """

    if psx is None:
        clf = LogisticRegression() if clf is None else clf
        psx = estimate_cv_predicted_probabilities(
            X=X,
            labels=y,
            clf=clf,
            cv_n_folds=n_folds,
            seed=seed,
        )

    conf_joint = compute_confident_joint(
        labels=y,
        pred_probs=psx
    )

    return conf_joint


def get_label_issues(
    X: np.array,
    y: np.array,
    n_folds: int = 5,
    clf: "BaseEstimator" = None,  # noqa: F821
    frac_noise: float = 1.0,
    num_to_remove_per_class: list = None,
    prune_method: str = 'prune_by_noise_rate',
    sorted_index_method: str = None,
    n_jobs: int = None,
    seed: int = None
) -> Tuple[np.array, np.array]:
    """
    Obtains samples from a dataset which can contain label issues, where issues could be samples that are incorrectly
    labeled, samples that could belong to more than one label category or some other issue. Internally, it uses the
    cleanlab library, which is based on confident learning:
    https://arxiv.org/pdf/1911.00068.pdf
    It also returns the confident joint matrix, which contains the counts of each one of the samples of a class that
    could belong to each one of the other classes


    Args:
        X (np.array): array of features of your dataset
        y (np.array): array of (noisy) labels of your dataset, i.e. some labels may be erroneous.
        n_folds (int): number of cross-validation folds used to compute out-of-sample probabilities for each sample
        clf (BaseEstimator): Classifier to calculate out-of-sample probabilities. If None, LogisticRegression is used.
        num_to_remove_per_class (list): Only used when `prune_method == 'prune_by_class'`. Indicates how many
            samples return for each class (returning the most likely mislabeled samples first).
        prune_method (str): Method used for filtering/pruning out the label issues.
            Possible Values: 'prune_by_class', 'prune_by_noise_rate', 'both', 'confident_learning' or 'predicted_neq_given'.
            Default value is 'prune_by_noise_rate'.
            You can check cleanlab documentation for further details (filter_by argument):
            https://docs.cleanlab.ai/master/cleanlab/filter.html#cleanlab.filter.find_label_issues
        sorted_index_method (str): Indicates how to sort the returned samples with issues. If `None` returns a boolean mask
            (`True` if example at index is label error). If not `None`, returns an array of the label error indices (instead
            of a boolean mask) where error indices are ordered.
            Possible values: None, 'self_confidence', 'normalized_margin', 'confidence_weighted_entropy'
            Default value is `None`.
            You can check cleanlab documentation for further details (return_indices_ranked_by argument):
            https://docs.cleanlab.ai/master/cleanlab/filter.html#cleanlab.filter.find_label_issues
        frac_noise (float): When frac_noise=1.0, return all "confident" estimated noise indices. Value in range (0,1]
            will return fewer samples. It only applies when `prune_method` is 'prune_by_class', 'prune_by_noise_rate' or 'both'
            Default value is `1.0`
            You can check cleanlab documentation for further details.
        n_jobs (int): Number of processing threads used by multiprocessing.
        seed (int): seed to use.

    Returns:
        returns the indexes of the samples that are detected as containing the issues and the confident joint matrix

    """

    clf = LogisticRegression() if clf is None else clf

    psx = estimate_cv_predicted_probabilities(
        X=X,
        labels=y,
        clf=clf,
        cv_n_folds=n_folds,
        seed=seed,
    )

    confident_joint = get_confident_joint(X, y, psx, n_folds, clf, seed)

    cl_label_errors = find_label_issues(
        labels=y,
        pred_probs=psx,
        confident_joint=confident_joint,
        frac_noise=frac_noise,
        num_to_remove_per_class=num_to_remove_per_class,
        filter_by=prune_method,
        return_indices_ranked_by=sorted_index_method,
        n_jobs=n_jobs
    )

    return cl_label_errors, confident_joint